
/*
 * $Id: CltXact.cc,v 1.1.1.10 1999/02/22 02:50:58 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <strstream.h>

#include "Clock.h"
#include "Connection.h"
#include "StatPing.h"
#include "StatPhaseMgr.h"
#include "SharedOpts.h"
#include "World.h"
#include "CltOpts.h"
#include "Client.h"
#include "CltXact.h"
#include "polyErrors.h"
#include "globals.h"


void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theRepSize = theConsumedSize = 0;
	theState = stNone;

	theRepHdr.reset();
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);

	theId.create();
	makeReq(theConn->theWrBuf);
	theState = stSpaceWaiting;
	noteWriteReady(-1); // kludge, sort of 
}

void CltXact::finish(int errNo) {
	stop(errNo);
	theOwner->noteXactDone(this);
}

void CltXact::stopStats() {
	TheStatPhaseMgr->xactRdSz(theConsumedSize, theLifeTime);
	Xaction::stopStats();
}

void CltXact::noteReadReady(int) {
	theConn->read();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theState == stHdrWaiting) {
		if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {

			// dump reply headers if needed
			if (TheOpts.doPrintReps)
				cout.write(theConn->theRdBuf.content(), theRepHdr.theHdrSize);

			if (theRepHdr.theContSize < 0) { // no content len
				finish(errNoCLen);
				return;
			}

			theRepSize = theRepHdr.theHdrSize + theRepHdr.theContSize;

			// note: read foreign replies, but
			// do not count foreign replies as misses, hits, etc.!
			if (theRepHdr.theXactId) {

				if (theOidRec.cachable != theRepHdr.isCachable)
					ReportError(errChbChange);

				theOidRec.mhit = !theId.myMutant(theRepHdr.theXactId);

				if (!theOidRec.ohit && theOidRec.mhit && !TheCltOpts.ignoreFalseHits)
					finish(errFalseHit);

				if (theOidRec.mhit)
					CallStat(ThePinger, realHit(theRepSize));
				else
					CallStat(ThePinger, realMiss(theRepSize));

				if (theOidRec.ohit)
					CallStat(ThePinger, idealHit(theRepSize));
				else
					CallStat(ThePinger, idealMiss(theRepSize));

				CallStat(ThePinger, chbObj(theRepSize, theRepHdr.isCachable));

				// hmm, foreigners are not noted?
				TheWorld.noteResponse(theOidRec);

				if (TheCltOpts.printFalseMisses && theOidRec.ohit && !theOidRec.mhit) {
					cout << "False Miss Oid: " << theOidRec.oid << endl;
					cout.write(theConn->theRdBuf.content(), theRepHdr.theHdrSize);
				}
			}

			theState = stBodyWaiting;
		} else
		if (theConn->theRdBuf.full()) {
			finish(errHugeHdr); // too big header
			return;
		}
	}

	if (theState == stBodyWaiting) {
		theConsumedSize += theConn->theRdBuf.contSize();
		theConn->theRdBuf.consumedAll();
		if (theConsumedSize == theRepSize) {
			// account for foreign replies here
			finish(theRepHdr.theXactId ? 0 : errForeignRep);
			return;
		}
	}

	if (theConn->exhausted()) { // premature eof
		finish(errPrematureEof);
		return;
	}
}

void CltXact::noteWriteReady(int) {

	Assert(theConn->theWrBuf.contSize() > 0);
	theConn->write();

	if (theConn->bad())
		finish(-1);
	else
	if (theConn->theWrBuf.empty()) {
		theConn->theWr.stop(this);
		theState = stHdrWaiting;
		theConn->theRd.start(this);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

void CltXact::makeReq(WrBuf &buf) {
	static int oid_off = Max(0, TheCltOpts.theWorld.theUniqStart - 1);
	const int oid = TheWorld.oid(oid_off, theOidRec);

	// XXX: strstream is not the most efficient implementation
	// XXX: can be further optimized with pre-build headers and Strings

	const char *reqStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	// request line prefix

	// make absolute and relative urls of the same length
	// so transparent folks do not get an advantage
	if (TheCltOpts.genAbsUrls)
		os << "GET http://";
	else
		os << "GET /http:/";

	const NetAddr &orig = TheCltOpts.theOrgAddr;
	os << orig.addrA();
	if (orig.port() != 80)
		os << ':' << orig.port();

	// url
	if (const char *url = TheWorld.url(oid)) {
		os << url; // preloaded
	} else {
		os << "/_" << oid << '/' << TheWorld.id() << '/' << TheWorld.type();

		if (TheCltOpts.theWorld.theUniqStart > 0)
			os << '/' << TheWorld.uid(oid_off);
	}

	os	<< " HTTP/1.0" << crlf;

	// headers
	os
		<< "Host: " << orig.addrA() << ':' << orig.port() << crlf
		<< (theOidRec.cachable ? "ETag: \"11\"" : "ETag: \"10\"") << crlf
		<< "Accept: */*" << crlf
		<< "X-Xact: " << theId << crlf
		<< crlf;

	oid_off++;
	buf.appended(os.tellp());

	// dump request
	if (TheOpts.doPrintReqs)
		cout.write(reqStart, buf.space()-reqStart);
}
