
/*
 * $Id: polysrv.cc,v 1.1.1.13 1999/02/22 06:19:50 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <new.h>

// IRIX demands _BSD_SIGNALS to be defined for signal()
// catching function declaration to have "int" as the first argument
#ifndef _BSD_SIGNALS
#define _BSD_SIGNALS
#endif
#include <signal.h>

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "Clock.h"
#include "Server.h"
#include "ListenQueue.h"
#include "StatPhaseMgr.h"
#include "Rnd.h"
#include "OLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "SrvOpts.h"

#include "gadgets.h"
#include "polyErrors.h"
#include "logTags.h"
#include "shutdown.h"
#include "globals.h"

static ListenQueue TheLQueue;

// runs one select loop cycle
static
int step(bool &idle) {
	FD_Set &rdSet = TheSelect->set(dirRead);
	FD_Set &wrSet = TheSelect->set(dirWrite);

	// default: use alarm timeout
	idle = false;
	Time tout = TheAlarmClock.timeLeft();
	Time *tptr = &tout;
	if (TheAlarmClock.clear()) { // no alarms
		if (TheSrvOpts.theTimeout >= 0) { // use global timeout (if any)
			tout = TheSrvOpts.theTimeout;
			idle = true;
		} else
			tptr = 0; // no global timeout
	}

	const int readyCount = TheSelect->select(tptr);
	CallStat(ThePinger, selected());

	if (readyCount > 0) {
		TheLQueue.begCycle(readyCount);
		for (int fd = TheSelect->maxFD(); fd >= 0; --fd) {
			if (rdSet.isReady(fd))
				rdSet[fd]->noteReadReady(fd);
			if (wrSet.isReady(fd))
				wrSet[fd]->noteWriteReady(fd);
			TheLQueue.tick(fd);
		}
		TheLQueue.endCycle();
	} else
	if (readyCount < 0 && errno != EINTR)
		FatalError(errno);

	return readyCount;
}

static
void configureLogs(int prec) {
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	if (TheOpts.theLogFileName)
		TheLog().stream(TheOpts.theLogFileName, new ofstream(TheOpts.theLogFileName));
	TheLog().resize(TheOpts.theLogSize);
	Comment.TheEchoLevel = TheOpts.theVerbLevel;

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

// set some general stuff and
// propogate cmd line options to corresponding objects
static
void configure() {

	configureLogs(2);

	atexit(&Shutdown);

	ErrorMgr::Configure(new PolyErrMgr);

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// quit nicely on some signals
	signal(SIGINT, &ShutdownSignal);

	// set resource limits
	const int fdLimit = SetMaxFD(GetMaxFD());
	if (fdLimit <= 0)
		FatalError(errZeroFDLimit);
	TheSelect->fdLimit(fdLimit);

	Socket::TheMaxLevel = Max(0, fdLimit - 10 - fdLimit/33);
	if (TheOpts.theFDLimit >= 0 && TheOpts.theFDLimit < Socket::TheMaxLevel)
		Socket::TheMaxLevel = TheOpts.theFDLimit;
	TheOpts.theFDLimit.set(Socket::TheMaxLevel);

	// set random seed for DistrOpts
	// note that distributions seeded with oids do not really require this
	TheOpts.passSeed(TheOpts.theRngSeed);
	TheSrvOpts.passSeed(TheOpts.theRngSeed);

	TheStatPhaseMgr.goal(TheOpts.theGoal);
}

int main(int argc, char *argv[]) {

	CmdLine cmd(Array<OptGrp*>() << &TheOpts << &TheSrvOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate() || !TheSrvOpts.validate())
		return -1;

	Clock::Update(false);

	configure();

	cmd.report();

	Comment(2) << "FDs: "
		<< GetCurFD() << " out of " << GetMaxFD()
		<< " FDs can be used; safeguard limit: " << Socket::TheMaxLevel 
		<< endc;

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheLog() << bege(lgRunId) << runId << ende;

	Clock::Update();
	Comment << "waiting for requests; goal: " << TheOpts.theGoal << endc;

	ThePinger.start();
	TheSoOftenPinger.start();

	Server TheServer;
	TheServer.configure(TheSrvOpts.thePort, TheOpts.theXactThinkTime.condDistr());
	TheServer.listenQueue(&TheLQueue);
	TheServer.start();

	while (!ShutdownNow()) {
		Clock::Update();
		ThePinger.checkpoint();
		TheSoOftenPinger.checkpoint();
		bool idle;
		if (!step(idle) && idle) {
			Comment << "was idle for at least " << TheSrvOpts.theTimeout << endc;
			ShutdownReason("incativity timeout");
		}
	}

	return 0;
}
